from os.path import exists, join, dirname
from os import symlink, mkdir


def init(conf, indir):
    for name in ('js', 'css', 'scss'):
        basedir = join(indir, conf.get(name, name))
        if not exists(basedir):
            continue

        # Project lib directory (e.g. myapp/js/lib)
        projpath = join(basedir, 'lib')
        # wq source directory (e.g. wq.app/js)
        wqpath = join(dirname(dirname(__file__)), name)

        if not exists(projpath):
            # e.g myapp/js/lib -> wq.app/js
            if name == "scss":
                # scss/lib should always be a folder
                mkdir(projpath)
            else:
                symlink(wqpath, projpath)
        if exists(join(wqpath, 'wq')) and not exists(join(projpath, 'wq')):
            # e.g myapp/js/lib/wq -> wq.app/js/wq
            symlink(join(wqpath, 'wq'), join(projpath, 'wq'))
        if name == "scss" and not exists(join(projpath, 'compass')):
            import pkg_resources
            compass = pkg_resources.resource_filename(
                'compass_stylesheets',
                join('stylesheets', 'compass'),
            )
            symlink(compass, join(projpath, 'compass'))



from __future__ import print_function

import os
import subprocess
import random
import json

from .collect import readfiles


def optimize(conf, indir, outdir):
    bfile = "rjsconf%s" % (random.random() * 10000)
    bjs = open(bfile, 'w')
    json.dump(conf, bjs)
    bjs.close()

    # Defer to r.js for actual processing
    print('#' * 20)
    print("Optimizing with r.js")
    rjs = os.path.dirname(__file__) + "/r.js"
    subprocess.call(["node", rjs, "-o", bfile])
    os.remove(bfile)
    os.remove(outdir + '/' + bfile)
    print("Optimization complete")
    print('#' * 20)


def scss(conf, indir=None):
    import scss
    import logging
    compiler = scss.Scss(scss_opts={'compress': 0})
    logging.getLogger("scss").addHandler(logging.StreamHandler())

    def compile(path, source):
        css = compiler.compile(source)
        outfile = open(path, 'w')
        outfile.write(css)
        outfile.close()

    if 'indir' in conf and 'outdir' in conf:
        files = readfiles(conf['indir'], "scss")
        scss.config.LOAD_PATHS = [
            conf['indir'],
            os.path.join(conf['indir'], 'lib'),
        ]
        for name, source in files.items():
            if isinstance(source, dict):
                continue
            path = "%s/%s.css" % (conf['outdir'], name)
            compile(path, source)
            print("%s compiled from %s/%s.scss" % (path, conf['indir'], name))


def mustache(conf, indir=None):
    import pystache
    template = conf.get("template", None)
    if template is None:
        return
    if os.sep in template or template.endswith(".html"):
        template = open(template).read()

    context = conf.get("context", {})
    if not isinstance(context, dict):
        path = context
        context = readfiles(path, "yaml", "yml")
        context.update(**readfiles(path, "json"))

    partials = conf.get("partials", {})
    if not isinstance(partials, dict):
        partials = readfiles(partials, "html")

    output = conf.get("output", "output.html")
    print("Generating %s from %s" % (output, conf['template']))
    renderer = pystache.Renderer(partials=partials)
    html = renderer.render(template, context)
    f = open(output, 'w')
    f.write(html)
    f.close()



from __future__ import print_function

import os
import json
try:
    import yaml
except:
    yaml = None

NEST = {
    'json': json,
    'yaml': yaml
}


def readfiles(basedir, ftype=None, fext=None):
    obj = {}
    if fext is None:
        fext = ftype

    for path, dirs, files in os.walk(basedir):
        if '.svn' in dirs:
            dirs.remove('.svn')
        o = obj
        if path == basedir:
            path = ""
        else:
            path = path[len(basedir) + 1:]
            apath = path.split(os.sep)
            for subdir in apath:
                o = o[subdir]
            path = os.sep.join(apath) + os.sep

        for filename in files:
            name, ext = os.path.splitext(filename)
            if ftype and ext != '.' + fext:
                continue

            fpath = path + name
            data = open(basedir + os.sep + fpath + ext)
            if NEST.get(ftype, None):
                try:
                    o[name] = NEST[ftype].load(data)
                except ValueError:
                    print("Could not parse %s!" % name)
                    raise
            else:
                o[name] = data.read()

        for name in dirs:
            o[name] = {}

    return obj


def collectjson(conf, directory):
    "Collect files and dump the result into a JSON object"
    obj = {}
    cur = os.getcwd()
    os.chdir(directory)

    for d in conf['paths']:
        obj.update(readfiles(d, conf['type'], conf.get('extension', None)))

    outfile = open(conf['output'], 'w')

    opts = dict([
        (str(key), value)
        for key, value in conf.get('json', {'indent': 4}).items()
    ])

    if 'jsonp' in conf:
        txt = json.dumps(obj, **opts)
        txt = '%s(%s);' % (conf['jsonp'], txt)
        outfile.write(txt)
    else:
        json.dump(obj, outfile, **opts)

    print('%s: %s objects collected from %s' % (
        conf['output'], len(obj), ', '.join(conf['paths'])
    ))

    outfile.close()
    os.chdir(cur)



from __future__ import print_function

import os


def setversion(conf, indir):
    filename = conf.get('filename', indir + '/version.txt')
    if (conf.get('version', None) is None):
        if os.path.exists(filename):
            version = open(filename).read().strip()
        else:
            version = ""
    else:
        version = conf['version']
        vtxt = open(filename, 'w')
        vtxt.write(version)
        vtxt.close()

    if 'jsout' in conf:
        # Update version.js
        vjs = open(indir + '/' + conf['jsout'], 'w')
        vjs.write(VERSIONJS_TMPL % version)
        vjs.close()
        print('%s: %s' % (conf['jsout'], version))
    else:
        print('Application version: %s' % version)

    return version

VERSIONJS_TMPL = """define(function(){return "%s";});"""



import json
import yaml
import sys

from .collect import collectjson
from .setversion import setversion
from .appcache import appcache
from .compilers import optimize, scss, mustache
from .init import init


class Builder(object):
    conf = None
    indir = None
    outdir = None
    version = ""

    COMMANDS = {
        'init': init,
        'collectjson': collectjson,
        'scss': scss,
        'mustache': mustache,
    }

    def __init__(self, version=None, config="wq.yml"):
        # Load configuration file
        try:
            self.conf = yaml.load(open(config))
        except (OSError, IOError):
            try:
                self.conf = json.load(open("app.build.json"))
            except (OSError, IOError):
                raise Exception("Could not find configuration file.")
            else:
                sys.stderr.write(
                    "Warning: Converted app.build.json to wq.yml\n"
                )
                yaml.dump(self.conf, open("wq.yml", 'w'))

        if 'optimize' not in self.conf:
            raise Exception("No optimize section in conf file!")

        # Determine input and output directories
        self.indir = self.conf['optimize']['appDir']
        self.outdir = self.conf['optimize']['dir']
        if version is not None:
            self.version = version

    def build(self):
        self.run('init')

        # Save version information
        if 'setversion' in self.conf or self.version != "":
            self.setversion()

        for command in ('init', 'collectjson', 'scss', 'mustache'):
            if command in self.conf:
                self.run(command)

        # Compile Javascript / CSS (using r.js)
        self.optimize()

        # Generate HTML5 Cache manifests
        if 'appcache' in self.conf:
            self.appcache()

    def run(self, command, directory=None, conf=None):
        if directory is None:
            directory = self.indir
        if conf is None:
            conf = self.conf.get(command, {})

        fn = self.COMMANDS[command]
        if isinstance(conf, dict):
            fn(conf, directory)
        else:
            for c in conf:
                fn(c, directory)

    def setversion(self, directory=None, conf=None):
        if directory is None:
            directory = self.indir
        if conf is None:
            conf = self.conf.get('setversion', {})
        if self.version != '':
            conf['version'] = self.version
        self.version = setversion(conf, directory)

    def optimize(self, conf=None):
        "Combine and optimize Javascript and CSS files"
        if conf is None:
            conf = self.conf.get('optimize', {})

        optimize(conf, self.indir, self.outdir)

    def appcache(self, conf=None):
        if conf is None:
            conf = self.conf.get('appcache', {})
        appcache(conf, self.indir, self.outdir, self.version)



from __future__ import print_function

import sys
from . import Builder


def usage():
    print("""Usage: wq [command] [version] [configfile]

Valid commands:
   appcache
   build
   collectjson
   init
   mustache
   optimize
   scss
   setversion
""")


def run():
    cmd = sys.argv[1] if len(sys.argv) > 1 else ""
    if len(sys.argv) > 2:
        args = sys.argv[2:]
    else:
        args = []
    if cmd == "":
        usage()
        exit()

    builder = Builder(*args)
    if cmd in builder.COMMANDS:
        builder.run(cmd)
    else:
        fn = getattr(builder, cmd, None)
        if fn is not None:
            fn()
        else:
            usage()



from __future__ import print_function

import os
import re


def appcache(conf, indir, outdir, version):

    # Open output files
    s_acpath = indir + '/' + conf['name']
    b_acpath = outdir + '/' + conf['name']
    s_ac = open(s_acpath, 'w')
    b_ac = open(b_acpath, 'w')

    # Start in source directory - read @imports from main CSS file
    s_css = [conf['css']]
    s_css.extend(_parse_css_urls(
        indir, conf['css'], conf.get('css-ignore', None)
    ))

    # Built CSS file contains image URLs from the @import-ed CSS files above
    images = _parse_css_urls(outdir, conf['css'], conf.get('css-ignore', None))

    # build.txt contains a list of built javascript files and their sources
    s_js, b_js = _parse_js_buildfile(outdir + '/build.txt')

    b_css = [conf['css']]

    # Collect path names and create appcaches
    cache = list(conf['cache'])
    source_cache = cache + s_js + s_css + images
    built_cache = cache + b_js + b_css + images

    network = list(conf['network'])
    fallback = list(conf['fallback'])

    s_ac.write(APPCACHE_TMPL % {
        'version':  '%s_dev' % version,
        'cache':    '\n'.join(source_cache),
        'network':  '\n'.join(network),
        'fallback': '\n'.join(fallback)
    })
    b_ac.write(APPCACHE_TMPL % {
        'version':  version,
        'cache':    '\n'.join(built_cache),
        'network':  '\n'.join(network),
        'fallback': '\n'.join(fallback)
    })
    s_ac.close()
    b_ac.close()
    print("%s: %s items" % (s_acpath, len(source_cache + network + fallback)))
    print("%s: %s items" % (b_acpath, len(built_cache + network + fallback)))


def _parse_js_buildfile(filename):
    text = open(filename).read()
    sources = []
    built = []
    for group in text[1:].split('\n\n'):
        files = group.split('\n')

        if files[0][-2:] != 'js':
            continue
        sources.extend(files[2:])
        built.append(files[0])

    return sources, built


def _parse_css_urls(path, filename, ignore=None):
    cur = os.getcwd()
    os.chdir(path)
    text = open(filename).read()
    base = os.path.dirname(filename)
    if ignore:
        ignore_re = re.compile(ignore)

    def parse_url(url):
        if url.startswith('data:') or url.startswith('#'):
            return None
        if url.startswith('http://') or url.startswith('https://'):
            return url
        if ignore and ignore_re.match(url):
            return None

        url = os.path.normpath(base + '/' + url)
        if os.sep != '/':
            url = url.replace(os.sep, '/')
        if ignore and ignore_re.match(url):
            return None
        return url

    urls = []
    for line in text.split('\n'):
        m = re.search(r"url\(['\" ]*(.+?)['\" ]*\)", line)
        if m:
            url = parse_url(m.group(1))
            if url:
                urls.append(url)
    os.chdir(cur)
    return urls

APPCACHE_TMPL = """CACHE MANIFEST

# Version %(version)s

CACHE:
%(cache)s

NETWORK:
%(network)s

FALLBACK:
%(fallback)s
"""
